--***************************************************************
--* File         : TB_WBSPIM.vhd                                *
--* Created      : time-stamp: <08/04/04 02:24:44 kawasan>      *
--* Modified     : Time-stamp: <08/05/16 02:13:23 kawasan>      *
--*								*
--* Test-Bench                                                  *
--*								*
--* Copyright 2008 Osamu Kawashima. all rights reserved		*
--***************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

use std.textio.all;
use ieee.std_logic_textio.all;

entity TB_WBSPIM is end;

architecture behavior of TB_WBSPIM is
    
	component WBSPIM
	generic (
		REG_CTRL_INIT   : std_logic_vector(7 downto 0) := x"00";
		REG_DIV_INIT    : std_logic_vector(7 downto 0) := x"00"
	);
	port (
		-- WISHBONE SLAVE interface:
		WB_CLK_I        : in    std_logic;
		WB_RST_I        : in    std_logic;
		WB_ADR_I        : in    std_logic_vector(7 downto 0);
		WB_DAT_I        : in    std_logic_vector(7 downto 0);
		WB_WE_I         : in    std_logic;
		WB_STB_I        : in    std_logic;
		WB_ACK_O        : out   std_logic;
		WB_DAT_O        : out   std_logic_vector(7 downto 0);
		WB_IRQ_O        : out   std_logic;

		-- Other port interface:
		SPI_SS_N        : out   std_logic;	-- Active Low
		SPI_SCLK        : out   std_logic;	-- Serial Clock
		SPI_MOSI        : out   std_logic;	-- Master-Out Slave-In
		SPI_MISO        : in    std_logic;	-- Master-In Slave-Out

		DONE            : out   std_logic
	);
	end component;

	constant ZERO32         : std_logic_vector(31 downto 0) := (others=>'0');
	constant ONE32          : std_logic_vector(31 downto 0) := (others=>'1');

	--constant REG_DIV_INIT   : std_logic_vector(7 downto 0) := x"00";	-- SYSCLK/2=24MHz
	--constant REG_DIV_INIT   : std_logic_vector(7 downto 0) := x"01";	-- SYSCLK/4=12MHz
	constant REG_DIV_INIT   : std_logic_vector(7 downto 0) := x"04";	-- SYSCLK/32=1.5MHz
	--constant REG_DIV_INIT   : std_logic_vector(7 downto 0) := x"05";	-- SYSCLK/64=750KHz

	constant WT1            : Time := 10 ns;
	constant WT_DELAY       : Time := 1 ns;
	constant WT_IDLE        : Time := 100 ns;
	constant CLK_CYCLE      : Time := 21 ns;	-- about 48MHz

	-- Address
	constant C_DATA         : std_logic_vector := x"00";
	constant C_CTRL         : std_logic_vector := x"01";
	constant C_DIV          : std_logic_vector := x"02";
	constant C_STAT         : std_logic_vector := x"03";

	signal WB_CLK_I         : std_logic := '0';
	signal WB_RST_I         : std_logic := '1';
	signal WB_ADR_I         : std_logic_vector(7 downto 0) := (others=>'0');
	signal WB_DAT_I         : std_logic_vector(7 downto 0) := (others=>'0');
	signal WB_WE_I          : std_logic := '0';
	signal WB_STB_I         : std_logic := '0';
	signal WB_ACK_O         : std_logic := '0';
	signal WB_DAT_O         : std_logic_vector(7 downto 0) := (others=>'0');
	signal WB_IRQ_O         : std_logic := '0';

	signal SPI_S_N          : std_logic := '1';
	signal SPI_CLK          : std_logic := '0';
	signal SPI_DO           : std_logic := '0';
	signal SPI_DI           : std_logic := '0';

	signal DONE             : std_logic := '0';

	signal rd_data          : std_logic_vector(7 downto 0) := (others=>'0');
	signal rd_dat           : std_logic_vector(7 downto 0) := (others=>'0');

	-- WB Read (8bit)
	procedure rd_ub(
		ADD_I		: in    std_logic_vector(7 downto 0);
		DAT_I		: in    std_logic_vector(7 downto 0);
	signal	ADD_O		: out   std_logic_vector(7 downto 0);
	signal	DAT_O		: out   std_logic_vector(7 downto 0);
	signal	STB		: out   std_logic;
	signal	WE		: out   std_logic
	) is
	begin
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		ADD_O	<= ADD_I;
		STB	<= '1';
		WE	<= '0';
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		DAT_O	<= DAT_I;
		wait for WT_DELAY;
		ADD_O	<= (others=>'0');
		STB	<= '0';
		WE	<= '0';
	end rd_ub;

	-- WB Write (8bit)
	procedure wr_ub(
		ADD_I		: in    std_logic_vector(7 downto 0);
		DAT_I		: in    std_logic_vector(7 downto 0);
	signal	ADD_O		: out   std_logic_vector(7 downto 0);
	signal	DAT_O		: out   std_logic_vector(7 downto 0);
	signal	STB		: out   std_logic;
	signal	WE		: out   std_logic
	) is
	begin
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		ADD_O	<= ADD_I;
		DAT_O	<= DAT_I;
		STB	<= '1';
		WE	<= '1';
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		ADD_O	<= (others=>'0');
		DAT_O	<= (others=>'0');
		STB	<= '0';
		WE	<= '0';
	end wr_ub;

begin
    
	uWBSPIM : WBSPIM
	generic map (
		REG_CTRL_INIT   => x"00",
		REG_DIV_INIT    => x"04"
	)
	port map (
		-- WISHBONE SLAVE interface:
		WB_CLK_I        => WB_CLK_I,
		WB_RST_I        => WB_RST_I,
		WB_ADR_I        => WB_ADR_I,
		WB_DAT_I        => WB_DAT_I,
		WB_WE_I         => WB_WE_I,
		WB_STB_I        => WB_STB_I,
		WB_ACK_O        => WB_ACK_O,
		WB_DAT_O        => WB_DAT_O,
		WB_IRQ_O        => WB_IRQ_O,

		-- Other port interface:
		SPI_SS_N        => SPI_S_N,
		SPI_SCLK        => SPI_CLK,
		SPI_MOSI        => SPI_DO,
		SPI_MISO        => SPI_DI,

		DONE            => DONE
	);

	process(WB_RST_I,WB_CLK_I)
	begin
		if WB_RST_I='1' then
			rd_dat <= (others=>'0');
		elsif WB_CLK_I'event and WB_CLK_I='1' then
			if WB_STB_I='1' and WB_WE_I='0' then
				rd_dat <= WB_DAT_O;
			else
				rd_dat <= rd_dat;
			end if;
		end if;
	end process;

	process begin
		WB_CLK_I <= '0';
		wait for CLK_CYCLE/2;
		WB_CLK_I <= '1';
		wait for CLK_CYCLE/2;
	end process;

	process
		file VEC        : text is in "spi.vec";
		variable L      : line;
		variable D_IN   : std_logic_vector(7 downto 0);
	begin
		readline (VEC, L);
		read (L, D_IN);

		for iCnt in 7 downto 0 loop
			wait until SPI_CLK'event and SPI_CLK='0';
			wait for WT_DELAY;
			if SPI_S_N='0' then
				SPI_DI <= D_IN(iCnt);
			end if;

			wait until SPI_CLK'event and SPI_CLK='1';
			wait for WT_DELAY;
		end loop;

		if endfile(VEC) then
			SPI_DI <= 'Z';
			wait;
		end if;
	end process;

	process
		variable ln         : line;
	begin
		-- Reset
		wait for WT_IDLE+WT_DELAY;
		WB_RST_I <= '0';
		wait for WT_IDLE;

		-- Init
		wr_ub(C_DIV, REG_DIV_INIT, WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);
		wr_ub(C_CTRL, x"03", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- SPI_SS_N=Active('0') & Irq
		wr_ub(C_DATA, x"55", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- Txd=0x55

		-- IRQ-Handler
		for iCnt in 1 to 8 loop
			wait until WB_IRQ_O'event and WB_IRQ_O='1';
			wait for WT_IDLE;
			rd_ub(C_STAT, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);
			rd_ub(C_DATA, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);

			write(ln, string'("C_DATA = "));
			hwrite(ln, rd_dat);
			writeline(output, ln);

			rd_ub(C_STAT, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);

			wr_ub(C_DATA, x"89", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- Txd=0x89
		end loop;

		-- End
		write(ln, string'("--- End of Test-Bench. ---"));
		writeline(output, ln);
		wait;
	end process;

end behavior;
