--***************************************************************
--* File         : WBSPIM.vhd                                   *
--* Created      : time-stamp: <08/04/05 02:43:32 kawasan>      *
--* Modified     : Time-stamp: <08/05/16 02:10:57 kawasan>      *
--*                                                             *
--* WISHBONE SPI Controller                                     *
--* Master-I/F                                                  *
--* MSB first                                                   *
--* Mode3 (http://elm-chan.org/docs/spi.html)                   *
--*                                                             *
--* Copyright 2008 Osamu Kawashima. all rights reserved         *
--***************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WBSPIM is
	generic (
		REG_CTRL_INIT   : std_logic_vector(7 downto 0) := x"00";
		REG_DIV_INIT    : std_logic_vector(7 downto 0) := x"01"
	);
	port (
		-- WISHBONE SLAVE interface:
		WB_CLK_I        : in    std_logic;
		WB_RST_I        : in    std_logic;
		WB_ADR_I        : in    std_logic_vector(7 downto 0);
		WB_DAT_I        : in    std_logic_vector(7 downto 0);
		WB_WE_I         : in    std_logic;
		WB_STB_I        : in    std_logic;
		WB_ACK_O        : out   std_logic;
		WB_DAT_O        : out   std_logic_vector(7 downto 0);
		WB_IRQ_O        : out   std_logic;

		-- Other port interface:
		SPI_SS_N        : out   std_logic;	-- Active Low
		SPI_SCLK        : out   std_logic;	-- Serial Clock
		SPI_MOSI        : out   std_logic;	-- Master-Out Slave-In
		SPI_MISO        : in    std_logic;	-- Master-In Slave-Out

		DONE            : out   std_logic
	);
end WBSPIM;

architecture RTL of WBSPIM is

	constant ZERO32         : std_logic_vector(31 downto 0) := (others=>'0');
	constant ONE32          : std_logic_vector(31 downto 0) := (others=>'1');

	-- Address
	constant C_DATA         : std_logic_vector := x"00";
	constant C_CTRL         : std_logic_vector := x"01";
	constant C_DIV          : std_logic_vector := x"02";
	constant C_STAT         : std_logic_vector := x"03";

	-- Bit
	constant B_DATA_IRQ_EN  : integer := 0;
	constant B_DATA_SPI_SS  : integer := 1;

	type STATE is (S_STANDBY, S_RUN, S_WAIT, S_LAST);
	signal next_state       : STATE;

	signal reg_ctrl         : std_logic_vector(1 downto 0);
	signal reg_div          : std_logic_vector(3 downto 0);

	signal dato_data        : std_logic_vector(WB_DAT_O'range);
	signal dato_ctrl        : std_logic_vector(WB_DAT_O'range);
	signal dato_div         : std_logic_vector(WB_DAT_O'range);
	signal dato_stat        : std_logic_vector(WB_DAT_O'range);

	signal clkdv_cnt        : std_logic_vector(WB_DAT_I'range);
	signal clkdv_cnt_max    : std_logic_vector(WB_DAT_I'range);
	signal clkdv_cnt_end    : std_logic;
	signal clkdv_event      : std_logic;
	signal clkdv_event0     : std_logic;
	signal clkdv_event1     : std_logic;

	signal spi_cnt          : std_logic_vector(3 downto 0);

	signal enable           : std_logic;
	signal enable_clk       : std_logic;
	signal wr_en            : std_logic;
	signal rd_en            : std_logic;
	signal irq_enable       : std_logic;
	signal irq              : std_logic;
	signal done_is          : std_logic;

	signal spi_dat_tmp      : std_logic_vector(WB_DAT_I'range);
	signal spi_dat_next     : std_logic_vector(WB_DAT_I'range);
	signal spi_dat_rx       : std_logic_vector(WB_DAT_O'range);
	signal spi_shift        : std_logic;
	signal spi_latch        : std_logic;
	signal spi_dat_latch    : std_logic;
	signal spi_ss           : std_logic;
	signal spi_sclk_is      : std_logic;
	signal spi_mosi_is      : std_logic;

begin

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' or enable='0' then
				clkdv_cnt <= clkdv_cnt_max;
			elsif clkdv_cnt_end='1' then
				clkdv_cnt <= (others=>'0');
			else
				clkdv_cnt <= clkdv_cnt + 1;
			end if;
		end if;
	end process;

	clkdv_cnt_end <= '1' when enable='1' and clkdv_cnt=clkdv_cnt_max else '0';

	process (reg_div)
	begin
		case reg_div is
			when x"0" =>
				clkdv_cnt_max <= x"00";		-- CLK/2

			when x"1" =>
				clkdv_cnt_max <= x"01";		-- CLK/4

			when x"2" =>
				clkdv_cnt_max <= x"03";		-- CLK/8

			when x"3" =>
				clkdv_cnt_max <= x"07";		-- CLK/16

			when x"4" =>
				clkdv_cnt_max <= x"0F";		-- CLK/32

			when x"5" =>
				clkdv_cnt_max <= x"1F";		-- CLK/64

			when x"6" =>
				clkdv_cnt_max <= x"3F";		-- CLK/128

			when x"7" =>
				clkdv_cnt_max <= x"7F";		-- CLK/256

			when x"8" =>
				clkdv_cnt_max <= x"FF";		-- CLK/512

			when others  =>
				clkdv_cnt_max <= x"FF";		-- CLK/512
		end case;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' or enable='0' then
				clkdv_event <= '0';
			elsif clkdv_cnt_end='1' then
				clkdv_event <= not clkdv_event;
			else
				clkdv_event <= clkdv_event;
			end if;
		end if;
	end process;

	clkdv_event0 <= '1' when clkdv_event='0' and clkdv_cnt_end='1' else '0';
	clkdv_event1 <= '1' when clkdv_event='1' and clkdv_cnt_end='1' else '0';

	spi_shift <= clkdv_event0;
	spi_latch <= clkdv_event1;

	wr_en <= '1' when WB_WE_I='1' and WB_STB_I='1' else '0';
	rd_en <= '1' when WB_WE_I='0' and WB_STB_I='1' else '0';

	spi_dat_next <= spi_dat_tmp(6 downto 0) & spi_dat_latch;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				spi_dat_tmp <= (others=>'0');
				spi_mosi_is <= '1';
			elsif WB_ADR_I=C_DATA and wr_en='1' then
				spi_dat_tmp <= WB_DAT_I;
				spi_mosi_is <= spi_mosi_is;
			elsif spi_shift='1' then
				spi_dat_tmp <= spi_dat_next;
				spi_mosi_is <= spi_dat_tmp(7);
			else
				spi_dat_tmp <= spi_dat_tmp;
				spi_mosi_is <= spi_mosi_is;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				spi_dat_latch <= '0';
			elsif spi_latch='1' then
				spi_dat_latch <= SPI_MISO;
			else
				spi_dat_latch <= spi_dat_latch;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' or enable_clk='0' then
				spi_sclk_is <= '0';
			elsif clkdv_event0='1' then
				spi_sclk_is <= '1';
			elsif clkdv_event1='1' then
				spi_sclk_is <= '0';
			else
				spi_sclk_is <= spi_sclk_is;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				reg_ctrl <= REG_CTRL_INIT(reg_ctrl'range);
			elsif WB_ADR_I=C_CTRL and wr_en='1' then
				reg_ctrl <= WB_DAT_I(reg_ctrl'range);
			else
				reg_ctrl <= reg_ctrl;
			end if;
		end if;
	end process;

	enable     <= '1' when next_state=S_RUN or next_state=S_WAIT else '0';
	enable_clk <= '1' when next_state=S_RUN else '0';
	irq_enable <= reg_ctrl(B_DATA_IRQ_EN);
	spi_ss     <= reg_ctrl(B_DATA_SPI_SS);

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				reg_div <= REG_DIV_INIT(reg_div'range);
			elsif WB_ADR_I=C_DIV and wr_en='1' then
				reg_div <= WB_DAT_I(reg_div'range);
			else
				reg_div <= reg_div;
			end if;
		end if;
	end process;

	SPI_SS_N <= not spi_ss;
	SPI_SCLK <= not spi_sclk_is;
	SPI_MOSI <= spi_mosi_is;

	-- State-Machine --
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				next_state <= S_STANDBY;
				spi_cnt <= x"8";
				spi_dat_rx <= (others=>'0');
			else
				case next_state is
					when S_STANDBY =>
						if WB_ADR_I=C_DATA and wr_en='1' then
							next_state <= S_RUN;
							spi_cnt <= x"8";
						else
							next_state <= S_STANDBY;
							spi_cnt <= spi_cnt;
						end if;
						spi_dat_rx <= spi_dat_rx;

					when S_RUN =>
						if spi_cnt=0 then
							next_state <= S_WAIT;
							spi_cnt <= spi_cnt;
							spi_dat_rx <= spi_dat_next;
						elsif clkdv_event1='1' then
							next_state <= S_RUN;
							spi_cnt <= spi_cnt - 1;
							spi_dat_rx <= spi_dat_rx;
						end if;

					when S_WAIT =>
						if clkdv_event0='1' then
							next_state <= S_LAST;
						else
							next_state <= S_WAIT;
						end if;
						spi_cnt <= spi_cnt;
						spi_dat_rx <= spi_dat_rx;

					when S_LAST =>
						next_state <= S_STANDBY;
						spi_dat_rx <= spi_dat_rx;
						spi_cnt <= spi_cnt;

					when others  =>
						next_state <= S_STANDBY;
						spi_cnt <= spi_cnt;
						spi_dat_rx <= spi_dat_rx;
				end case;
			end if;
		end if;
	end process;
	-- State-Machine --

	process(WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' or (WB_ADR_I=C_DATA and rd_en='1') then
				irq <= '0';
			elsif next_state=S_LAST then
				irq <= '1';
			else
				irq <= irq;
			end if;
		end if;
	end process;

	WB_IRQ_O <= irq and irq_enable;
	WB_ACK_O <= WB_STB_I;

	dato_data  <= spi_dat_rx;
	dato_ctrl  <= ZERO32(WB_DAT_O'left downto reg_ctrl'left+1) & reg_ctrl(reg_ctrl'range);
	dato_div   <= ZERO32(WB_DAT_O'left downto reg_div'left+1) & reg_div(reg_div'range);
	dato_stat  <= ZERO32(WB_DAT_O'left downto 1) & irq;

	WB_DAT_O <= dato_data when WB_ADR_I=C_DATA else 
		    dato_ctrl when WB_ADR_I=C_CTRL else 
	            dato_div  when WB_ADR_I=C_DIV else 
	            dato_stat when WB_ADR_I=C_STAT else 
		    (others=>'0');

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' or next_state=S_LAST then
				done_is <= '1';
			elsif WB_ADR_I=C_DATA and wr_en='1' then
				done_is <= '0';
			else
				done_is <= done_is;
			end if;
		end if;
	end process;

	DONE <= done_is;

end RTL;
